@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2019-2023 by Sukchan Lee <acetcom@gmail.com>
+ * Copyright (C) 2019-2024 by Sukchan Lee <acetcom@gmail.com>
  *
  * This file is part of Open5GS.
  *
@@ -25,6 +25,15 @@ static int discover_handler(
 static void handle_nf_discover_search_result(
         OpenAPI_search_result_t *SearchResult);
 
+/**
+ * Handles NF registration in NRF. Validates the PLMN-ID against configured
+ * serving PLMN-IDs and registers the NF instance if valid.
+ *
+ * @param nf_instance The NF instance being registered.
+ * @param stream The SBI stream for communication.
+ * @param recvmsg The received SBI message.
+ * @return true if registration is successful; otherwise, false.
+ */
 bool nrf_nnrf_handle_nf_register(ogs_sbi_nf_instance_t *nf_instance,
         ogs_sbi_stream_t *stream, ogs_sbi_message_t *recvmsg)
 {
@@ -33,6 +42,10 @@ bool nrf_nnrf_handle_nf_register(ogs_sbi_nf_instance_t *nf_instance,
 
     OpenAPI_nf_profile_t *NFProfile = NULL;
 
+    OpenAPI_lnode_t *node = NULL;
+    bool plmn_valid = false;
+    int i;
+
     ogs_assert(nf_instance);
     ogs_assert(stream);
     ogs_assert(recvmsg);
@@ -73,6 +86,41 @@ bool nrf_nnrf_handle_nf_register(ogs_sbi_nf_instance_t *nf_instance,
         return false;
     }
 
+    /* Validate the PLMN-ID against configured serving PLMN-IDs */
+    if (NFProfile->plmn_list) {
+        /* Set PLMN status to invalid */
+        plmn_valid = false;
+
+        if (ogs_local_conf()->num_of_serving_plmn_id > 0 && NFProfile->plmn_list) {
+            OpenAPI_list_for_each(NFProfile->plmn_list, node) {
+                OpenAPI_plmn_id_t *PlmnId = node->data;
+                if (PlmnId == NULL) {
+                    continue;
+                }
+                for (i = 0; i < ogs_local_conf()->num_of_serving_plmn_id; i++) {
+                    if (ogs_sbi_compare_plmn_list(
+                                &ogs_local_conf()->serving_plmn_id[i],
+                                PlmnId) == true) {
+                        plmn_valid = true;
+                        break;
+                    }
+                }
+                if (plmn_valid) {
+                    break;
+                }
+            }
+        }
+
+        /* Reject the registration if PLMN-ID is invalid */
+        if (!plmn_valid) {
+            ogs_error("PLMN-ID in NFProfile is not allowed");
+            ogs_assert(true == ogs_sbi_server_send_error(
+                stream, OGS_SBI_HTTP_STATUS_BAD_REQUEST, recvmsg,
+                "PLMN-ID not allowed", NULL, NULL));
+            return false;
+        }
+    }
+
     ogs_nnrf_nfm_handle_nf_profile(nf_instance, NFProfile);
 
     ogs_sbi_client_associate(nf_instance);
@@ -146,7 +194,6 @@ bool nrf_nnrf_handle_nf_register(ogs_sbi_nf_instance_t *nf_instance,
         if (ogs_local_conf()->num_of_serving_plmn_id &&
                 NFProfile->plmn_list == NULL) {
             OpenAPI_list_t *PlmnIdList = NULL;
-            int i;
 
             PlmnIdList = OpenAPI_list_create();
             ogs_assert(PlmnIdList);
@@ -208,6 +255,10 @@ bool nrf_nnrf_handle_nf_update(ogs_sbi_nf_instance_t *nf_instance,
     ogs_assert(stream);
     ogs_assert(recvmsg);
 
+    cJSON *plmn_array = NULL, *plmn_item = NULL;
+    bool plmn_valid = false;
+    int i;
+
     SWITCH(recvmsg->h.method)
     CASE(OGS_SBI_HTTP_METHOD_PUT)
         return nrf_nnrf_handle_nf_register(
@@ -224,14 +275,14 @@ bool nrf_nnrf_handle_nf_update(ogs_sbi_nf_instance_t *nf_instance,
             return false;
         }
 
+        /* Iterate through the PatchItemList */
         OpenAPI_list_for_each(PatchItemList, node) {
             OpenAPI_patch_item_t *patch_item = node->data;
             if (!patch_item) {
                 ogs_error("No PatchItem");
-                ogs_assert(true ==
-                    ogs_sbi_server_send_error(stream,
-                        OGS_SBI_HTTP_STATUS_BAD_REQUEST,
-                        recvmsg, "No PatchItem", NULL, NULL));
+                ogs_assert(true == ogs_sbi_server_send_error(
+                    stream, OGS_SBI_HTTP_STATUS_BAD_REQUEST, recvmsg,
+                    "No PatchItem", NULL, NULL));
                 return false;
             }
 
@@ -246,6 +297,102 @@ bool nrf_nnrf_handle_nf_update(ogs_sbi_nf_instance_t *nf_instance,
                 break;
             CASE(OGS_SBI_PATCH_PATH_LOAD)
                 break;
+            CASE(OGS_SBI_PATCH_PATH_PLMN_LIST)
+                /* Ensure the value is not null and is a valid JSON array */
+                if (patch_item->value && patch_item->value->json) {
+                    /* Set PLMN status to invalid */
+                    plmn_valid = false;
+
+                    plmn_array = patch_item->value->json;
+                    if (!cJSON_IsArray(plmn_array)) {
+                        ogs_error("Value for /plmnList is not a JSON array");
+                        ogs_assert(true == ogs_sbi_server_send_error(
+                            stream, OGS_SBI_HTTP_STATUS_BAD_REQUEST, recvmsg,
+                            "Invalid value for /plmnList", NULL, NULL));
+                        return false;
+                    }
+
+                    /* Clear existing PLMN data in nf_instance */
+                    memset(nf_instance->plmn_id, 0,
+                            sizeof(nf_instance->plmn_id));
+                    nf_instance->num_of_plmn_id = 0;
+
+                    /* Iterate through the JSON array of PLMN IDs */
+                    cJSON_ArrayForEach(plmn_item, plmn_array) {
+                        OpenAPI_plmn_id_t plmn_id;
+                        memset(&plmn_id, 0, sizeof(plmn_id));
+
+                        if (nf_instance->num_of_plmn_id >=
+                                OGS_ARRAY_SIZE(nf_instance->plmn_id)) {
+                            ogs_error("Exceeded maximum number of PLMN IDs");
+                            ogs_assert(true == ogs_sbi_server_send_error(
+                                stream, OGS_SBI_HTTP_STATUS_BAD_REQUEST,
+                                recvmsg,
+                                "Too many PLMN IDs", NULL, NULL));
+                            return false;
+                        }
+
+                        /* Parse the PLMN item */
+                        plmn_id.mcc = cJSON_GetObjectItem(plmn_item, "mcc")
+                                           ? cJSON_GetStringValue(
+                                                 cJSON_GetObjectItem(
+                                                     plmn_item, "mcc"))
+                                           : NULL;
+                        plmn_id.mnc = cJSON_GetObjectItem(plmn_item, "mnc")
+                                           ? cJSON_GetStringValue(
+                                                 cJSON_GetObjectItem(
+                                                     plmn_item, "mnc"))
+                                           : NULL;
+
+                        if (!plmn_id.mcc || !plmn_id.mnc) {
+                            ogs_error(
+                                "Invalid PLMN item in /plmnList update");
+                            ogs_assert(true ==
+                                       ogs_sbi_server_send_error(
+                                           stream,
+                                           OGS_SBI_HTTP_STATUS_BAD_REQUEST,
+                                           recvmsg,
+                                           "Invalid PLMN item", NULL,
+                                           NULL));
+                            return false;
+                        }
+
+                        /*
+                         * Convert OpenAPI_plmn_id_t to ogs_plmn_id_t
+                         * and store in nf_instance
+                         */
+                        ogs_sbi_parse_plmn_id(
+                                &nf_instance->
+                                    plmn_id[nf_instance->num_of_plmn_id],
+                                &plmn_id);
+                        nf_instance->num_of_plmn_id++;
+
+                        /* Compare with the serving PLMN list */
+                        for (i = 0;
+                             i < ogs_local_conf()->num_of_serving_plmn_id;
+                             i++) {
+                            if (ogs_sbi_compare_plmn_list(
+                                        &ogs_local_conf()->serving_plmn_id[i],
+                                        &plmn_id) == true) {
+                                plmn_valid = true;
+                                break;
+                            }
+                        }
+                        if (plmn_valid) {
+                            break;
+                        }
+                    }
+
+                    /* Reject the update if PLMN-ID is invalid */
+                    if (!plmn_valid) {
+                        ogs_error("PLMN-ID in NFProfile update is not allowed");
+                        ogs_assert(true == ogs_sbi_server_send_error(
+                            stream, OGS_SBI_HTTP_STATUS_BAD_REQUEST, recvmsg,
+                            "PLMN-ID not allowed", NULL, NULL));
+                        return false;
+                    }
+                }
+                break;
             DEFAULT
                 ogs_error("Unknown PatchItem.Path [%s]", patch_item->path);
             END
@@ -352,7 +499,8 @@ bool nrf_nnrf_handle_nf_status_subscribe(
 
     if (SubscriptionData->requester_features) {
         subscription_data->requester_features =
-            ogs_uint64_from_string(SubscriptionData->requester_features);
+            ogs_uint64_from_string_hexadecimal(
+                    SubscriptionData->requester_features);
 
         /* No need to send SubscriptionData->requester_features to the NF */
         ogs_free(SubscriptionData->requester_features);
@@ -386,6 +534,9 @@ bool nrf_nnrf_handle_nf_status_subscribe(
         else if (SubscrCond->service_name)
             subscription_data->subscr_cond.service_name =
                 ogs_strdup(SubscrCond->service_name);
+        else if (SubscrCond->nf_instance_id)
+            subscription_data->subscr_cond.nf_instance_id = 
+                ogs_strdup(SubscrCond->nf_instance_id);
         else {
             ogs_error("No SubscrCond");
             ogs_sbi_subscription_data_remove(subscription_data);
@@ -424,13 +575,13 @@ bool nrf_nnrf_handle_nf_status_subscribe(
     /*
      * The NRF validity is initially set in configuration.
      */
-    subscription_data->time.validity_duration =
-            ogs_local_conf()->time.subscription.validity_duration;
+    subscription_data->validity_duration =
+        ogs_time_from_sec(
+                ogs_local_conf()->time.subscription.validity_duration);
 
-    if (subscription_data->time.validity_duration) {
+    if (subscription_data->validity_duration) {
         SubscriptionData->validity_time = ogs_sbi_localtime_string(
-            ogs_time_now() + ogs_time_from_sec(
-                subscription_data->time.validity_duration));
+            ogs_time_now() + subscription_data->validity_duration);
         ogs_assert(SubscriptionData->validity_time);
 
         if (!subscription_data->t_validity) {
@@ -440,13 +591,16 @@ bool nrf_nnrf_handle_nf_status_subscribe(
             ogs_assert(subscription_data->t_validity);
         }
         ogs_timer_start(subscription_data->t_validity,
-                ogs_time_from_sec(subscription_data->time.validity_duration));
+                subscription_data->validity_duration);
     }
 
-    ogs_info("[%s] Subscription created until %s [validity_duration:%d]",
+    ogs_info("[%s] Subscription created until %s "
+            "[duration:%lld,validity:%d.%06d]",
             subscription_data->id,
             SubscriptionData->validity_time,
-            subscription_data->time.validity_duration);
+            (long long)subscription_data->validity_duration,
+            (int)ogs_time_sec(subscription_data->validity_duration),
+            (int)ogs_time_usec(subscription_data->validity_duration));
 
     /* Location */
     server = ogs_sbi_server_from_stream(stream);
@@ -487,6 +641,8 @@ bool nrf_nnrf_handle_nf_status_update(
 
     ogs_sbi_subscription_data_t *subscription_data = NULL;
 
+    ogs_time_t time, validity;
+
     ogs_assert(stream);
     ogs_assert(recvmsg);
 
@@ -553,73 +709,87 @@ bool nrf_nnrf_handle_nf_status_update(
         END
     }
 
-    if (validity_time) {
-        ogs_time_t time, validity;
-
-        if (ogs_sbi_time_from_string(&time, validity_time) == false) {
-            ogs_error("[%s] Subscription updated until %s [parser error]",
-                    subscription_data->id, validity_time);
-            goto end;
-        }
-
-        validity = time - ogs_time_now();
-        if (validity < 0) {
-            ogs_error("[%s] Subscription updated until %s [validity:%d.%06d]",
-                    subscription_data->id, validity_time,
-                    (int)ogs_time_sec(validity), (int)ogs_time_usec(validity));
-            goto end;
-        }
-
-        /*
-         * The NRF validity is updated if NF sent the PATCH Request.
-         */
-        subscription_data->time.validity_duration =
-                    OGS_SBI_VALIDITY_SEC(validity);
-
-        if (!subscription_data->t_validity) {
-            subscription_data->t_validity =
-                ogs_timer_add(ogs_app()->timer_mgr,
-                    nrf_timer_subscription_validity, subscription_data);
-            ogs_assert(subscription_data->t_validity);
-        }
-        ogs_timer_start(subscription_data->t_validity,
-                ogs_time_from_sec(subscription_data->time.validity_duration));
-
-        ogs_info("[%s] Subscription updated until %s "
-                "[duration:%d,validity:%d.%06d]",
-                subscription_data->id, validity_time,
-                subscription_data->time.validity_duration,
-                (int)ogs_time_sec(validity), (int)ogs_time_usec(validity));
-
-        /* To send SubscriptionData to the NF */
-        memset(&sendmsg, 0, sizeof(sendmsg));
-        sendmsg.SubscriptionData = &SubscriptionData;
-
-        /* Mandatory */
-        SubscriptionData.nf_status_notification_uri =
-            subscription_data->notification_uri;
-
-        /* Validity Time */
-        SubscriptionData.validity_time = ogs_sbi_localtime_string(
-            ogs_time_now() + ogs_time_from_sec(
-                subscription_data->time.validity_duration));
-        ogs_assert(SubscriptionData.validity_time);
-
-        response = ogs_sbi_build_response(&sendmsg, OGS_SBI_HTTP_STATUS_OK);
-        ogs_assert(response);
-        ogs_assert(true == ogs_sbi_server_send_response(stream, response));
-
-        ogs_free(SubscriptionData.validity_time);
-
-        return true;
+    if (!validity_time) {
+        ogs_error("[%s] No validityTime", subscription_data->id);
+        ogs_assert(true ==
+            ogs_sbi_server_send_error(
+                stream, OGS_SBI_HTTP_STATUS_BAD_REQUEST,
+                recvmsg, "No validityTime", subscription_data->id,
+                NULL));
+        return false;
     }
 
-end:
-    response = ogs_sbi_build_response(
-            recvmsg, OGS_SBI_HTTP_STATUS_NO_CONTENT);
+    if (ogs_sbi_time_from_string(&time, validity_time) == false) {
+        ogs_error("[%s] Subscription updated until %s [parser error]",
+                subscription_data->id, validity_time);
+        ogs_assert(true ==
+            ogs_sbi_server_send_error(
+                stream, OGS_SBI_HTTP_STATUS_BAD_REQUEST,
+                recvmsg, "parse error", subscription_data->id,
+                validity_time));
+        return false;
+    }
+
+    validity = time - ogs_time_now();
+    if (validity < 0) {
+        ogs_error("[%s] Subscription updated until %s [validity:%d.%06d]",
+                subscription_data->id, validity_time,
+                (int)ogs_time_sec(validity), (int)ogs_time_usec(validity));
+        ogs_assert(true ==
+            ogs_sbi_server_send_error(
+                stream, OGS_SBI_HTTP_STATUS_BAD_REQUEST,
+                recvmsg, "invalid validity", subscription_data->id,
+                validity_time));
+        return false;
+    }
+
+    /*
+     * The NRF validity is updated if NF sent the PATCH Request.
+     */
+    subscription_data->validity_duration =
+        /* Normalize seconds */
+        ogs_time_from_sec(ogs_time_to_sec(validity));
+
+    if (!subscription_data->t_validity) {
+        subscription_data->t_validity =
+            ogs_timer_add(ogs_app()->timer_mgr,
+                nrf_timer_subscription_validity, subscription_data);
+        ogs_assert(subscription_data->t_validity);
+    }
+    ogs_timer_start(subscription_data->t_validity,
+            subscription_data->validity_duration);
+
+    ogs_info("[%s] Subscription updated until %s "
+            "[duration:%lld,validity:%d.%06d]",
+            subscription_data->id, validity_time,
+            (long long)subscription_data->validity_duration,
+            (int)ogs_time_sec(subscription_data->validity_duration),
+            (int)ogs_time_usec(subscription_data->validity_duration));
+
+    /* To send SubscriptionData to the NF */
+    memset(&sendmsg, 0, sizeof(sendmsg));
+
+#if 0 /* Use HTTP_STATUS_NO_CONTENT(204) instead of HTTP_STATUS_OK(200) */
+    sendmsg.SubscriptionData = &SubscriptionData;
+
+    /* Mandatory */
+    SubscriptionData.nf_status_notification_uri =
+        subscription_data->notification_uri;
+
+    /* Validity Time */
+    SubscriptionData.validity_time = ogs_sbi_localtime_string(
+        ogs_time_now() + subscription_data->validity_duration);
+    ogs_assert(SubscriptionData.validity_time);
+
+    response = ogs_sbi_build_response(&sendmsg, OGS_SBI_HTTP_STATUS_OK);
+#else
+    response = ogs_sbi_build_response(&sendmsg, OGS_SBI_HTTP_STATUS_NO_CONTENT);
+#endif
     ogs_assert(response);
     ogs_assert(true == ogs_sbi_server_send_response(stream, response));
 
+    ogs_free(SubscriptionData.validity_time);
+
     return true;
 }
 
@@ -835,12 +1005,12 @@ bool nrf_nnrf_handle_nf_discover(
                             &discovery_option->tai.plmn_id),
                         discovery_option->tai.tac.v);
         }
-        if (discovery_option->target_guami) {
+        if (discovery_option->guami_presence) {
             ogs_debug("guami[PLMN_ID:%06x,AMF_ID:%x]",
                         ogs_plmn_id_hexdump(
-                            &discovery_option->target_guami->plmn_id),
+                            &discovery_option->guami.plmn_id),
                         ogs_amf_id_hexdump(
-                            &discovery_option->target_guami->amf_id));
+                            &discovery_option->guami.amf_id));
         }
         if (discovery_option->num_of_target_plmn_list) {
             for (i = 0; i < discovery_option->num_of_target_plmn_list; i++)
@@ -858,6 +1028,9 @@ bool nrf_nnrf_handle_nf_discover(
                     ogs_plmn_id_mnc(
                         &discovery_option->requester_plmn_list[i]));
         }
+        if (discovery_option->hnrf_uri) {
+            ogs_debug("hnrf_uri[%s]", discovery_option->hnrf_uri);
+        }
         if (discovery_option->requester_features) {
             ogs_debug("requester-features[0x%llx]",
                 (long long)discovery_option->requester_features);
@@ -892,11 +1065,12 @@ bool nrf_nnrf_handle_nf_discover(
         if (recvmsg->param.limit && i >= recvmsg->param.limit)
             break;
 
-        ogs_debug("[%s:%d] NF-Discovered [NF-Type:%s,NF-Status:%s,"
-                "IPv4:%d,IPv6:%d]", nf_instance->id, i,
+        ogs_info("[%s:%d] NF-Discovered [NF-Type:%s,NF-Status:%s,"
+                "IPv4:%d,IPv6:%d,FQDN:%s]", nf_instance->id, i,
                 OpenAPI_nf_type_ToString(nf_instance->nf_type),
                 OpenAPI_nf_status_ToString(nf_instance->nf_status),
-                nf_instance->num_of_ipv4, nf_instance->num_of_ipv6);
+                nf_instance->num_of_ipv4, nf_instance->num_of_ipv6,
+                nf_instance->fqdn ? nf_instance->fqdn : "NULL");
 
         NFProfile = ogs_nnrf_nfm_build_nf_profile(
                 nf_instance, NULL, discovery_option,
@@ -963,6 +1137,16 @@ bool nrf_nnrf_handle_nf_discover(
                         nf_instance, discovery_option) == false)
                 continue;
 
+            /*
+             * NRFs in Home PLMN are categorized into those with and without
+             * hnrf-uri. Since Open5GS includes hnrf-uri in the Discovery
+             * message, NRFs must be differentiated based on the presence or
+             * absence of hnrf-uri, so the following match routine is used.
+             */
+            if (ogs_sbi_discovery_option_hnrf_uri_is_matched(
+                        nf_instance, discovery_option) == false)
+                continue;
+
             break;
         }
 
@@ -984,10 +1168,63 @@ bool nrf_nnrf_handle_nf_discover(
             nf_instance->num_of_plmn_id =
                 discovery_option->num_of_target_plmn_list;
 
-            nf_instance->fqdn = ogs_nrf_fqdn_from_plmn_id(nf_instance->plmn_id);
-            ogs_assert(nf_instance->fqdn);
+            /*
+             * NRFs with a Home PLMN can use the target_plmn information
+             * to extract the FQDN. However, the standards documentation
+             * describes hnrf-uri, so if hnrf-uri is included in discovery,
+             * we implement it so that the FQDN of an NRF with a Home PLMN
+             * is created using hnrf-uri.
+             */
+            if (discovery_option->hnrf_uri) {
+                OpenAPI_uri_scheme_e scheme = OpenAPI_uri_scheme_NULL;
+                char *fqdn = NULL;
+                uint16_t fqdn_port = 0;
+                ogs_sockaddr_t *addr = NULL, *addr6 = NULL;
 
+                rc = ogs_sbi_getaddr_from_uri(
+                        &scheme, &fqdn, &fqdn_port, &addr, &addr6,
+                        discovery_option->hnrf_uri);
+                if (rc == false || scheme == OpenAPI_uri_scheme_NULL)
+                    ogs_error("Invalid URL [%s]", request->h.uri);
+                else {
+            /*
+             * If there is an hnrf-uri, this value creates an nf-instance->fqdn,
+             * which in turn creates a client->fqdn from the hnrf-uri.
+             */
+                    nf_instance->hnrf_uri =
+                        ogs_strdup(discovery_option->hnrf_uri);
+                    nf_instance->fqdn = ogs_strdup(fqdn);
+
+                    ogs_free(fqdn);
+                    ogs_freeaddrinfo(addr);
+                    ogs_freeaddrinfo(addr6);
+                }
+            }
+
+            if (!nf_instance->fqdn) {
+                nf_instance->fqdn =
+                    ogs_nrf_fqdn_from_plmn_id(nf_instance->plmn_id);
+                ogs_assert(nf_instance->fqdn);
+            }
+
+            /*
+             * At this stage, the client->fqdn is set and this is
+             * passed through to the Target-apiRoot to the Home PLMN.
+             *
+             * As mentioned above, if there is an hnrf-uri, this value creates
+             * an nf-instance->fqdn, which in turn creates a client->fqdn
+             * from the hnrf-uri.
+             */
             ogs_sbi_client_associate(nf_instance);
+
+            ogs_info("New NRF [fqdn:%s, hnrf_uri:%s]",
+                    nf_instance->fqdn ? nf_instance->fqdn : "NULL",
+                    nf_instance->hnrf_uri ? nf_instance->hnrf_uri : "NULL");
+
+        } else {
+            ogs_info("NRF Found [fqdn:%s, hnrf_uri:%s]",
+                    nf_instance->fqdn ? nf_instance->fqdn : "NULL",
+                    nf_instance->hnrf_uri ? nf_instance->hnrf_uri : "NULL");
         }
 
         client = NF_INSTANCE_CLIENT(nf_instance);
@@ -995,7 +1232,7 @@ bool nrf_nnrf_handle_nf_discover(
 
         /*
          * TS29.510
-         * 5.3.2.4 Service Discovery in a different PLMN
+         * 5.3.2.2.3 Service Discovery in a different PLMN
          *
          * Then, steps 1-2 in Figure 5.3.2.2.3-1 are executed,
          * between the NRF in the Serving PLMN and the NRF in the Home PLMN.
@@ -1003,6 +1240,7 @@ bool nrf_nnrf_handle_nf_discover(
          * in the query parameter of the URI is not required.
          */
         discovery_option->num_of_target_plmn_list = 0;
+        ogs_sbi_discovery_option_clear_hnrf_uri(discovery_option);
 
         assoc = nrf_assoc_add(stream);
         if (!assoc) {
